"""
Create vector grids from references or hardcode
"""

# TODO: Implement grid functions
# TODO: create_grid, create_grid_points, create_grid_linestrings

def create_grid():
    """ Not implemented."""
    raise Exception("Not implemented")
